
import asyncio
from multiprocessing import cpu_count
import multiprocessing
from typing import TYPE_CHECKING, Callable, Optional, Union
import threading
import time
import sys, os, subprocess
import logging
import datetime
import platform
from copy import copy
from queue import Empty

from rocketry._base import RedBase
from rocketry.core.condition import BaseCondition, AlwaysFalse
from rocketry.core.task import Task
from rocketry.exc import SchedulerRestart, SchedulerExit
from rocketry.core.hook import _Hooker

if TYPE_CHECKING:
    from rocketry import Session

class Scheduler(RedBase):
    """Multiprocessing scheduler

    Parameters
    ----------
    session : rocketry.session.Session, optional
        Session object containing tasks,
        parameters and settings, 
        by default None
    max_processes : int, optional
        Maximum number of processes
        allowed to be started, 
        by default number of CPUs
    tasks_as_daemon : bool, optional
        Whether process tasks are run
        as daemon (if not specified in
        the task) or not, by default True
    timeout : str, optional
        Timeout of each task if not specified
        in the task itself. Must be timedelta
        string, by default "30 minutes"
    parameters : [type], optional
        Parameters of the session.
        Can also be passed directly to the 
        session, by default None
    logger : [type], optional
        [description], by default None
    name : str, optional
        Name of the scheduler, deprecated, by default None
    restarting : str, optional
        How the scheduler is restarted if 
        Restart exception is raised, by default 
        "replace"
    instant_shutdown : bool, optional
        Whether the scheduler tries to shut down
        as quickly as possible or wait till all
        the tasks have finished, by default False

    Attributes
    ----------
    session : rocketry.session.Session
        Session for which the scheduler is
        for. One session has only one 
        scheduler.
    """
    session: 'Session'

    def __init__(self, session=None,
                logger=None, name:str=None):

        # Other
        self.session = session or self.session

        self.name = name if name is not None else id(self) #! TODO Is this needed?
        self.logger = logger


        # Controlling runtime (used by scheduler.disabled)
        self._flag_enabled = threading.Event()
        self._flag_shutdown = threading.Event()
        self._flag_force_exit = threading.Event()
        self._flag_restart = threading.Event()
        self._flag_enabled.set() # Not on hold by default

        # is_alive is used by testing whether the scheduler is 
        # still running or not
        self.is_alive = None

        self._log_queue = multiprocessing.Queue(-1)

    def _register_instance(self):
        self.session.scheduler = self

    @property
    def tasks(self):

        #! TODO: Is this needed?
        tasks = self.session.get_tasks()
        # There may be extra rare situation that priority is not in the task
        # for short period if it is being modified thus we use getattr
        return sorted(tasks, key=lambda task: getattr(task, "priority", 0), reverse=True)

    def __call__(self):
        return self.run()

    def run(self):
        return asyncio.run(self.serve())

    async def serve(self):
        """Start and run the scheduler. Will block till the end of the scheduling
        session."""
        # Unsetting some flags
        self._flag_shutdown.clear()
        self._flag_restart.clear()
        self._flag_enabled.set()

        self.is_alive = True
        exception = None
        try:
            await self.startup()

            while not self.check_shut_cond(self.session.config.shut_cond):
                await self._hibernate()
                if self._flag_shutdown.is_set():
                    break
                elif self._flag_restart.is_set():
                    raise SchedulerRestart()

                await self.run_cycle()

                # self.maintain()
        except SystemExit as exc:
            self.logger.info('Shutting down...', extra={"action": "shutdown"})
            exception = exc

        except SchedulerExit as exc:
            self.logger.info('Shutdown called. Shutting down...', exc_info=True, extra={"action": "shutdown"})
            exception = exc

        except SchedulerRestart as exc:
            self.logger.info('Restart called. Shutting down...', exc_info=True, extra={"action": "shutdown"})
            exception = exc

        except KeyboardInterrupt as exc:
            self.logger.info('Interupted. Shutting down...', exc_info=True, extra={"action": "shutdown"})
            exception = exc

        except Exception as exc:
            self.logger.critical('Fatal error encountered. Shutting down...', exc_info=True, extra={"action": "crash"})
            exception = exc
            raise
        else:
            self.logger.info('Purpose completed. Shutting down...', extra={"action": "shutdown"})
        finally:
            await self.shut_down(exception=exception)

    async def run_cycle(self):
        """Run one round of tasks.
        
        Each task is inspected and in case their starting condition
        is fulfilled, they are run.  A task can be running once at 
        any given time (in other words, multiple paraller execution 
        of a single task is not supported at the moment). Tasks that 
        are running but their termination condition is fulfilled are 
        terminated.
        """
        tasks = self.tasks
        self.logger.debug(f"Beginning cycle with {len(tasks)} tasks...", extra={"action": "run"})

        # Running hooks
        hooker = _Hooker(self.session.hooks.scheduler_cycle)
        hooker.prerun(self)

        for task in tasks:
            with task.lock:
                self.handle_logs()
                if task.on_startup or task.on_shutdown:
                    # Startup or shutdown tasks are not run in main sequence
                    pass
                elif self._flag_enabled.is_set() and self.is_task_runnable(task):
                    # Run the actual task
                    await self.run_task(task)
                    # Reset force_run as a run has forced
                    task.force_run = False
                elif self.is_timeouted(task):
                    # Terminate the task
                    await self.terminate_task(task, reason="timeout")
                elif self.is_out_of_condition(task):
                    # Terminate the task
                    await self.terminate_task(task)

        # Running hooks
        hooker.postrun()
        
        self.n_cycles += 1

    def check_shut_cond(self, cond: Optional[BaseCondition]) -> bool:
        # Note that failure in scheduler shut_cond always crashes the system
        if cond is None:
            return False
        return cond.observe(scheduler=self, session=self.session)

    def check_task_cond(self, task:Task):
        try:
            return task.is_runnable()
        except:
            if not self.session.config.silence_cond_check:
                raise
            return False

    async def run_task(self, task:Task, *args, **kwargs):
        """Run a given task"""
        start_time = datetime.datetime.fromtimestamp(time.time())

        try:
            await task.start_async(log_queue=self._log_queue)
        except (SchedulerRestart, SchedulerExit) as exc:
            raise 
        except Exception as exc:
            if not self.session.config.silence_task_prerun:
                raise
        else:
            exception = None
            status = "success"

    async def terminate_all(self, reason:str=None):
        """Terminate all running tasks."""
        for task in self.tasks:
            if task.is_alive():
                await self.terminate_task(task, reason=reason)

    async def terminate_task(self, task, reason=None):
        """Terminate a given task."""
        self.logger.debug(f"Terminating task '{task.name}'")
        is_threaded = hasattr(task, "_thread")
        is_multiprocessed = hasattr(task, "_process")
        if task.is_alive_as_thread():
            # We can only kindly ask the thread to...
            # get the fuck out please.
            task._thread_terminate.set()

        elif task.is_alive_as_process():
            task._process.terminate()
            # Waiting till the termination is finished. 
            # Otherwise may try to terminate it many times as the process is alive for a brief moment
            task._process.join() 
            task.log_termination(reason=reason)

            # Resetting attr force_termination
            task.force_termination = False
        elif task.is_alive_as_async():
            task._async_task.cancel()
            try:
                await task._async_task
            except asyncio.CancelledError:
                task.log_termination()
        else:
            # The process/thread probably just died after the check
            pass

    def is_timeouted(self, task):
        """Check if the task is timeouted."""
        #! TODO: Can this be put to the Task?
        if task.permanent_task:
            # Task is meant to be on all the time thus no reason to terminate due to timeout
            return False
        elif not hasattr(task, "_thread") and not hasattr(task, "_process"):
            # Task running on the main process
            # cannot be left running
            return False
        elif not task.is_alive():
            return False

        timeout = (
            task.timeout if task.timeout is not None
            else self.session.config.timeout
        )
        
        if timeout is None:
            return False
        run_duration = datetime.datetime.fromtimestamp(time.time()) - task.get_last_run()
        return run_duration > timeout

    def is_task_runnable(self, task:Task):
        """Inspect whether the task should be run."""
        #! TODO: Can this be put to the Task?
        execution = task.get_execution()
        is_condition = self.check_task_cond(task)
        if execution == "process":
            is_not_running = not task.is_alive()
            has_free_processors = self.has_free_processors()
            return is_not_running and has_free_processors and is_condition
        elif execution == "main":
            return is_condition
        elif execution == "thread":
            is_not_running = not task.is_alive()
            return is_not_running and is_condition
        elif execution == "async":
            is_not_running = not task.is_alive()
            return is_not_running and is_condition
        else:
            raise NotImplementedError(task.execution)

    def is_out_of_condition(self, task:Task):
        """Inspect whether the task should be terminated."""
        #! TODO: Can this be put to the Task?
        if not task.is_alive():
            # NOTE:
            # Task running on the main process
            # cannot be left running
            return False

        elif task.force_termination:
            return True

        else:
            return task.is_terminable()
            
    def handle_logs(self):
        """Handle the status queue and carries the logging on their behalf."""
        # TODO: This could be maybe done in the tasks
        queue = self._log_queue
        while True:
            try:
                record = queue.get(block=False)
            except Empty:
                break
            else:
                self.logger.debug(f"Inserting record for '{record.task_name}' ({record.action})")
                task = self.session.get_task(record.task_name)
                if record.action == "fail":
                    # There is a caveat in logging 
                    # https://github.com/python/cpython/blame/fad6af2744c0b022568f7f4a8afc93fed056d4db/Lib/logging/handlers.py#L1383 
                    # https://bugs.python.org/issue34334

                    # The traceback/exception info is no longer in record.exc_info/record.exc_text 
                    # and it has been formatted to record.message/record.msg
                    # This means we have to rely that message really contains
                    # the full traceback

                    record.exc_info = record.exc_text
                    record.exc_text = record.exc_text
                    if record.exc_text is not None and record.exc_text not in record.message:
                        record.message = record.message + "\n" + record.message
                elif record.action == "success":
                    # Take the return value from the record and delete
                    # Note that record has attr __return__ only if task running as process
                    return_value = record.__return__
                    task._handle_return(return_value)
                    del record.__return__
                
                task.log_record(record)

    async def _hibernate(self):
        """Go to sleep and wake up when next task can be executed."""
        delay = self.session.config.cycle_sleep
        if delay is not None:
            await asyncio.sleep(delay)

    async def startup(self):
        """Start up the scheduler.
        
        Starting up includes setting up attributes and
        running tasks that have ``on_startup`` as ``True``."""
        #self.setup_listener()
        self.logger.info(f"Starting up...", extra={"action": "setup"})
        hooker = _Hooker(self.session.hooks.scheduler_startup)
        hooker.prerun(self)

        self.n_cycles = 0
        self.startup_time = datetime.datetime.fromtimestamp(time.time())

        self.logger.info(f"Beginning startup sequence...")
        for task in self.tasks:
            if task.on_startup:
                if isinstance(task.start_cond, AlwaysFalse) and not task.disabled: 
                    # Make sure the tasks run if start_cond not set
                    task.force_run = True

                if self.is_task_runnable(task):
                    await self.run_task(task)

        hooker.postrun()
        self.logger.info(f"Setup complete.")

    def has_free_processors(self) -> bool:
        """Whether the Scheduler has free processors to
        allocate more tasks."""
        return self.n_alive <= self.session.config.max_process_count

    @property
    def n_alive(self) -> int:
        """Count of tasks that are alive."""
        return sum(task.is_alive() for task in self.tasks)
        
    async def _shut_down_tasks(self, traceback=None, exception=None):
        non_fatal_excs = (SchedulerRestart,) # Exceptions that are allowed to have graceful exit
        wait_for_finish = (
            not self.session.config.instant_shutdown 
            and (exception is None or isinstance(exception, non_fatal_excs))
        ) and not self._flag_force_exit.is_set()
        if wait_for_finish:
            try:
                # Gracefully shut down (allow remaining tasks to finish)
                while self.n_alive:
                    #time.sleep(self.min_sleep)

                    await self._hibernate() # This is the time async tasks can continue

                    self.handle_logs()
                    for task in self.tasks:
                        if task.permanent_task:
                            # Would never "finish" anyways
                            await self.terminate_task(task)
                        elif self.is_timeouted(task):
                            # Terminate the task
                            await self.terminate_task(task, reason="timeout")
                        elif self.is_out_of_condition(task):
                            # Terminate the task
                            await self.terminate_task(task)
            except Exception as exc:
                # Fuck it, terminate all
                await self._shut_down_tasks(exception=exc)
                return
            else:
                self.handle_logs()
        else:
            await self.terminate_all(reason="shutdown")

    async def wait_task_alive(self):
        """Wait till all, especially threading tasks, are finished."""
        while self.n_alive > 0:
            await self._hibernate()

    async def shut_down(self, traceback=None, exception=None):
        """Shut down the scheduler.
        
        Shutting down includes running tasks that have 
        ``on_shutdown`` as ``True``, handling the 
        shutting down of already running tasks and 
        restarting the scheduler in case restart was
        called.
        
        If non fatal exception was raised and ``instant_shutdown``
        is ``False``, remaining running tasks are waited to finish.
        Else all the tasks are terminated. If ``instant_shutdown``
        is ``True``, the scheduler won't wait for the terminated 
        tasks to finish their termination.
        """
        
        self.logger.info(f"Beginning shutdown sequence...")
        hooker = _Hooker(self.session.hooks.scheduler_shutdown)
        hooker.prerun(self)

        # Make sure the tasks run if start_cond not set
        for task in self.tasks:
            if task.on_shutdown:

                if isinstance(task.start_cond, AlwaysFalse) and not task.disabled: 
                    # Make sure the tasks run if start_cond not set
                    task.force_run = True

                if self.is_task_runnable(task):
                    await self.run_task(task)

        self.logger.info(f"Shutting down tasks...")
        await self._shut_down_tasks(traceback, exception)

        await self.wait_task_alive() # Wait till all tasks' threads and processes are dead

        # Running hooks
        hooker.postrun()

        self.is_alive = False
        self.logger.info(f"Shutdown completed. Good bye.")
        if isinstance(exception, SchedulerRestart):
            # Clean up finished, restart is finally
            # possible
            await self._restart()

    async def _restart(self):
        """Restart the scheduler by creating a new process
        on the temporary run script where the scheduler's is
        process is started.
        """
        # https://stackoverflow.com/a/35874988
        self.logger.debug(f"Restarting...", extra={"action": "restart"})
        python = sys.executable

        restarting = self.session.config.restarting
        if restarting == "replace":
            os.execl(python, python, *sys.argv)
            # After this, no code will be run. It all died :(
        elif restarting == "relaunch":
            # Relaunch the process
            subprocess.Popen([python, *sys.argv], shell=False, close_fds=True)
        elif restarting == "fresh":
            # Relaunch the process in new window
            if platform.system() == "Windows":
                subprocess.Popen([python, *sys.argv], shell=False, close_fds=True, creationflags=subprocess.CREATE_NEW_CONSOLE)
            else:
                # Linux does not have CREATE_NEW_CONSOLE creation flag but shell=True is pretty close
                subprocess.Popen([python, *sys.argv], shell=True, close_fds=True)
        elif restarting == "recall":
            # Mostly useful for testing.
            # Restart by calling the self.__call__ again 
            await asyncio.create_task(self.serve()) 
        else:
            raise ValueError(f"Invalid restaring: {restarting}")

# System control
    @property
    def on_hold(self):
        """bool: If True, the scheduler won't execute new tasks
        till this is set False. Useful to halt task execution in
        a controller task."""
        return not self._flag_enabled.is_set()

    @on_hold.setter
    def on_hold(self, value):
        if value:
            self._flag_enabled.clear()
        else:
            self._flag_enabled.set()

    def set_shut_down(self):
        """Shut down the scheduler. Useful to shut down the 
        scheduler in a controller task."""
        self.on_hold = False # In case was set to wait
        self._flag_shutdown.set()

# Logging
    @property
    def logger(self):
        return self._logger

    @logger.setter
    def logger(self, logger):
        basename = self.session.config.scheduler_logger_basename
        if logger is None:
            # Get class logger (default logger)
            logger = logging.getLogger(basename)

        if not logger.name.startswith(basename):
            raise ValueError(f"Logger name must start with '{basename}' as session finds loggers with names")

        # TODO: Use TaskAdapter to relay the scheduler name?
        self._logger = logger




import re
from typing import Union

from rocketry.core.condition import BaseCondition
from rocketry.args import Session

class IsEnv(BaseCondition):
    """Condition checks whether session parameter `env`
    has the given value. 

    Parameters
    ----------
    env : str
        The environment to be set in order to the condition
        to be ``True``.

    Examples
    --------
    >>> from rocketry.conditions import IsEnv
    >>> is_prod = IsEnv("prod")

    >>> # Correct environment
    >>> from rocketry import session
    >>> session.env = 'prod'
    >>> bool(is_prod)
    True

    >>> # Incorrect environment
    >>> session.env = 'dev'
    >>> bool(is_prod)
    False
    """
    __parsers__ = {re.compile(r"env '(?P<env>.+)'"): "__init__"}

    def __init__(self, env):
        self.env = env
    
    def get_state(self, session=Session()):
        return session.parameters.get("env", None) == self.env

class ParamExists(BaseCondition):
    """Condition to check whether parameter(s) (and their values)
    exists from ``session.parameters``.

    Parameters
    ----------
    *args : iterable of strings
        Names of the parameters expected to be 
        found from the session (in order the 
        condition to be True)
    **kwargs : dict
        Names of the parameters and their values 
        expected from the session (in order the 
        condition to be True)

    Examples
    --------
    >>> from rocketry.conditions import ParamExists
    >>> condition = ParamExists("z", x=1, y=2)
    
    >>> # Parameters found
    >>> from rocketry import session
    >>> session.parameters = {"x": 1, "y": 2, "z": 3, "k": 4}
    >>> bool(condition)
    True

    >>> # Missing parameter(s)
    >>> session.parameters = {"x": 1}
    >>> bool(condition)
    False
    """
    __parsers__ = {
        re.compile(r"param '(?P<l>.+)' exists"): "_from_list",
        re.compile(r"param '(?P<key>.+)' is '(?P<value>.+)'"): "_from_key_value",
    }

    param_keys:dict
    param_vals:tuple

    def __init__(self, *args, **kwargs):
        self.param_keys = args
        self.param_values = kwargs
    
    def __bool__(self):
        params = self.session.parameters
        for key in self.param_keys:
            if key not in params:
                return False
        for key, val in self.param_values.items():
            if key not in params:
                return False
            elif params[key] != val:
                return False
        # Passed all test
        return True

    @classmethod
    def _from_list(cls, l:Union[tuple, list]):
        return cls(*l)
    
    @classmethod
    def _from_key_value(cls, key:str, value):
        return cls(**{key: value})



import pytest

from rocketry.args import Private, Return
from rocketry import Scheduler
from rocketry.conditions.scheduler import SchedulerCycles
from rocketry.core import parameters
from rocketry.tasks import FuncTask
from rocketry.conditions import TaskStarted
from rocketry.args import FuncArg


def func_with_arg(value):
    pass

def func_x_with_return():
    return "x"

def func_x_with_arg(myparam):
    assert myparam == "x"


@pytest.mark.parametrize("execution", ["main", "thread", "process"])
def test_normal(session, execution):

    task_return = FuncTask(
        func_x_with_return, 
        name="return task",
        start_cond="~has started",
        execution=execution,
        force_run=True
    )
    task = FuncTask(
        func_x_with_arg, 
        name="a task",
        start_cond="after task 'return task'",
        parameters={"myparam": Return('return task')},
        execution=execution
    )

    assert task.status is None

    session.config.shut_cond = TaskStarted(task="a task") >= 1
    session.start()

    assert dict(session.returns) == {task_return: "x", task: None}
    assert "success" == task_return.status
    assert "success" == task.status

@pytest.mark.parametrize("execution", ["main", "thread", "process"])
def test_normal_pass_task(session, execution):

    task_return = FuncTask(
        func_x_with_return, 
        name="return task",
        start_cond="~has started",
        execution=execution,
        force_run=True
    )
    task = FuncTask(
        func_x_with_arg, 
        name="a task",
        start_cond="after task 'return task'",
        parameters={"myparam": Return(task_return)},
        execution=execution
    )

    assert task.status is None

    session.config.shut_cond = TaskStarted(task="a task") >= 1
    session.start()

    assert dict(session.returns) == {task_return: "x", task: None}
    assert "success" == task_return.status
    assert "success" == task.status

@pytest.mark.parametrize("execution", ["main", "thread", "process"])
def test_normal_pass_func(session, execution):
    # This use case is for using the app pattern
    @FuncTask(
        name="return task",
        start_cond="~has started",
        execution="main",
        force_run=True
    )
    def func_with_arg_local():
        return 'x'

    task = FuncTask(
        func_x_with_arg, 
        name="a task",
        start_cond="after task 'return task'",
        parameters={"myparam": Return(func_with_arg_local)},
        execution=execution
    )

    assert task.status is None

    session.config.shut_cond = (TaskStarted(task="a task") >= 1) | (SchedulerCycles() >= 5)
    session.start()

    task_return = session["return task"]
    assert dict(session.returns) == {task_return: "x", task: None}
    assert "success" == task_return.status
    assert "success" == task.status

@pytest.mark.parametrize("execution", ["main", "thread", "process"])
def test_missing(session, execution):
    session.config.silence_task_prerun = True # Default in prod
    
    task = FuncTask(
        func_with_arg, 
        parameters={"myparam": Return('return task')},
        name="a task",  
        force_run=True,
        execution=execution
    )
 
    assert task.status is None
    session.config.shut_cond = TaskStarted(task="a task") >= 1
    session.start()

    assert "fail" == task.status

@pytest.mark.parametrize("execution", ["main", "thread", "process"])
def test_default(session, execution):
    task = FuncTask(
        func_with_arg, 
        name="return task", 
    )
    task = FuncTask(
        func_x_with_arg, 
        parameters={"myparam": Return('return task', default="x")},
        name="a task", 
        execution=execution, 
        force_run=True,
    )

    assert task.status is None
    session.config.shut_cond = TaskStarted(task="a task") >= 1
    session.start()

    assert "success" == task.status



import logging
from textwrap import dedent
import pytest

from redbird.logging import RepoHandler
from redbird.repos import MemoryRepo

from rocketry.log.log_record import LogRecord
from rocketry.tasks import CodeTask
from rocketry.core import Scheduler
from rocketry.conditions import TaskStarted

def test_construct(session):
    task = CodeTask(code=dedent("""
        def main():
            ...
        main()
        """), start_cond="daily", name="mytask")
    assert task.name == "mytask"
    assert "mytask" in session

def test_construct_missing_name(session):
    with pytest.raises(ValueError):
        task = CodeTask(code=dedent("""
            def main():
                ...
            main()
            """))

@pytest.mark.parametrize('execution', ['main', 'thread', 'process'])
def test_run_success(session, execution):
    
    task = CodeTask(code=dedent("""
        def main():
            return 'myvalue'

        return_value = main()
        """), execution=execution, name="mytask")
    task.force_run = True

    session.config.shut_cond = TaskStarted(task='mytask') >= 1
    session.start()
    assert task.status == 'success'
    assert session.returns[task] == 'myvalue'

@pytest.mark.parametrize('execution', ['main', 'thread', 'process'])
def test_run_success_parametrize(session, execution):
    
    task = CodeTask(code=dedent("""
        def main(param):
            return 'myvalue' + param

        return_value = main(myparam)
        """), name="mytask", execution=execution, parameters={'myparam': ' + myparam'})
    task.force_run = True

    session.config.shut_cond = TaskStarted(task='mytask') >= 1
    session.start()

    assert task.status == 'success'
    assert session.returns[task] == 'myvalue + myparam'

@pytest.mark.parametrize('execution', ['main', 'thread', 'process'])
def test_run_fail(session, execution):
    
    task_logger = logging.getLogger(session.config.task_logger_basename)
    task_logger.handlers = [
        RepoHandler(repo=MemoryRepo(model=LogRecord))
    ]

    task = CodeTask(code=dedent("""
        def main():
            raise RuntimeError('Failed')

        return_value = main()
        """), execution=execution, name="mytask")
    task.force_run = True

    session.config.shut_cond = TaskStarted(task='mytask') >= 1
    session.start()
    
    assert task.status == 'fail'

    records = list(map(lambda e: e.dict(exclude={'created'}), session.get_task_log()))
    record_fail = [r for r in records if r['action'] == 'fail'][0]
    assert 'File "<string>", line 5, in <module>\n  File "<string>", line 3, in main\nRuntimeError: Failed' in record_fail['exc_text']





import logging
from pathlib import Path
import tempfile
from typing import List, Optional, Pattern, Union
import warnings

from redbird import BaseRepo
from redbird.logging import RepoHandler
from redbird.repos import MemoryRepo, CSVFileRepo
from rocketry.log.log_record import LogRecord

from rocketry.tasks import FuncTask, CommandTask
from rocketry.conditions import FuncCond
from rocketry.parameters import FuncParam
from rocketry import Session

class _AppMixin:

    def task(self, start_cond=None, name=None, *, command=None, path=None, **kwargs):
        "Create a task"

        kwargs['session'] = self.session
        kwargs['start_cond'] = start_cond
        kwargs['name'] = name

        if command is not None:
            return CommandTask(command=command, **kwargs)
        elif path is not None:
            # Non-wrapped FuncTask
            return FuncTask(path=path, **kwargs)
        else:
            return FuncTask(name_include_module=False, _name_template='{func_name}', **kwargs)

    def param(self, name:Optional[str]=None):
        "Set one session parameter (decorator)"
        return FuncParam(name, session=self.session)


class Rocketry(_AppMixin):
    """Rocketry scheduling application"""

    def __init__(self, session:Session=None, logger_repo:Optional[BaseRepo]=None, execution=None, **kwargs):

        self.session = session if session is not None else Session(**kwargs)

        logger = logging.getLogger(self.session.config.task_logger_basename)
        logger.setLevel(logging.INFO)

        if execution is not None:
            self.session.config.task_execution = execution

        self._set_logger_with_repo(logger_repo)

    def run(self, debug=False):
        "Run the scheduler"
        self.session.config.debug = debug
        self.session.set_as_default()
        self.session.start()

    async def serve(self, debug=False):
        "Run the scheduler"
        self.session.config.debug = debug
        self.session.set_as_default()
        await self.session.serve()

    def cond(self, syntax: Union[str, Pattern, List[Union[str, Pattern]]]=None):
        "Create a condition (decorator)"
        return FuncCond(syntax=syntax, session=self.session, decor_return_func=False)

    def params(self, **kwargs):
        "Set session parameters"
        self.session.parameters.update(kwargs)

    def set_logger(self):
        warnings.warn((
            "set_logger is deprecated and will be removed in the future. " 
            "Please set the logger using logging.getLogger"
        ), DeprecationWarning)
        def wrapper(func):
            func(self._get_task_logger())
        return wrapper

    def _get_task_logger(self):
        logger_name = self.session.config.task_logger_basename
        return logging.getLogger(logger_name)

    def _set_logger_with_repo(self, repo):
        if isinstance(repo, str):
            self._get_repo(repo)
        elif repo is None:
            repo = MemoryRepo(model=LogRecord)
        logger = self._get_task_logger()
        logger.handlers.insert(0, RepoHandler(repo=repo))
    
    def _get_repo(self, repo:str):
        if repo == "memory":
            return MemoryRepo(model=LogRecord)
        elif repo == "csv":
            filepath = Path(tempfile.gettempdir()) / "rocketry.csv"
            return CSVFileRepo(filename=filepath, model=LogRecord)
        else:
            raise NotImplementedError(f"Repo creation for {repo} not implemented")




from pathlib import Path
import types

import pytest

from rocketry.tasks import FuncTask
from rocketry.conditions import AlwaysFalse, AlwaysTrue, DependSuccess
from rocketry.parse.utils import ParserError

def myfunc(): ...

@pytest.mark.parametrize("execution", ["main", "thread", "process"])
def test_construct(tmpdir, session, execution):

    # Going to tempdir to dump the log files there
    with tmpdir.as_cwd() as old_dir:
        if execution == "process":
            with pytest.raises(AttributeError):
                # Unpicklable function (cannot use process)
                task = FuncTask(lambda : None, execution=execution)
        else:
            task = FuncTask(lambda : None, execution=execution)
            
        # This should always be picklable
        task = FuncTask(myfunc, execution=execution)
        assert not task.is_delayed()
        assert task.status is None

@pytest.mark.parametrize("execution", ["main", "thread", "process"])
def test_construct_callable_class(tmpdir, session, execution):
    class MyClass:
        def __call__(self):
            pass

    # Going to tempdir to dump the log files there
    with tmpdir.as_cwd() as old_dir:
        # This should always be picklable
        task = FuncTask(MyClass(), execution=execution)
        assert not task.is_delayed()
        assert task.status is None
        assert task.name.endswith("test_construct:MyClass")

def test_description(session):
    "Test Task.description is correctly set (uses the func if missing)"

    # Test no description
    task = FuncTask(name="no desc 1")
    @task
    def myfunc():
        ...
    assert task.description is None

    task = FuncTask(lambda x: x, name="no desc 2", execution="main")
    assert task.description is None


     # Test description from doc (decorated)
    task = FuncTask(name="has desc 1")
    @task
    def myfunc():
        "This is func"
    assert task.description == "This is func"

    # Test description from doc (normal)
    def myfunc():
        "This is func"
    task = FuncTask(myfunc, name="has desc 2")
    assert task.description == "This is func"

    # Test the description is respected if doc is found
    def myfunc():
        "This is func"
    task = FuncTask(myfunc, name="has desc 3", description="But this is preferred")
    assert task.description == "But this is preferred"

    # Test the description is respected if doc missing
    def myfunc():
        ...
    task = FuncTask(myfunc, name="has desc 4", description="This is func")
    assert task.description == "This is func"

@pytest.mark.parametrize("execution", ["main", "thread", "process"])
def test_construct_delayed(tmpdir, session, execution):

    # Going to tempdir to dump the log files there
    with tmpdir.as_cwd() as old_dir:
        task = FuncTask(func_name="myfunc", path="myfile.py", execution=execution)
        assert task.status is None
        assert task.is_delayed()
        assert task.func_name == "myfunc"
        assert task.path == Path("myfile.py")
        assert task.func is None

def test_construct_decorate(tmpdir, session):
    # Going to tempdir to dump the log files there
    with tmpdir.as_cwd() as old_dir:


        @FuncTask(start_cond=AlwaysTrue(), name="mytask", execution="main")
        def do_stuff():
            pass
        
        assert isinstance(do_stuff, types.FunctionType)

        do_stuff_task = session["mytask"]
        assert isinstance(do_stuff_task, FuncTask)
        assert do_stuff_task.status is None
        assert do_stuff_task.start_cond == AlwaysTrue()
        assert do_stuff_task.name == "mytask"

        assert {do_stuff_task} == session.tasks 

def test_construct_decorate_minimal(tmpdir, session):
    """This is an exception when FuncTask returns itself 
    (__init__ cannot return anything else)"""
    # Going to tempdir to dump the log files there
    orig_default_exec = session.config.task_execution
    session.config.task_execution = 'main'
    try:
        with tmpdir.as_cwd() as old_dir:

            @FuncTask
            def do_stuff():
                pass

            assert isinstance(do_stuff, FuncTask)
            assert do_stuff.status is None
            assert do_stuff.start_cond == AlwaysFalse()
            assert do_stuff.name.endswith(":do_stuff")

            assert {do_stuff} == session.tasks
    finally:
        session.config.task_execution = orig_default_exec

def test_construct_decorate_default_name(tmpdir, session):
    # Going to tempdir to dump the log files there
    with tmpdir.as_cwd() as old_dir:


        @FuncTask(start_cond=AlwaysTrue(), execution="main")
        def do_stuff():
            pass
        
        assert isinstance(do_stuff, types.FunctionType)
        do_stuff_task = list(session.tasks)[-1]
        assert isinstance(do_stuff_task, FuncTask)
        assert do_stuff_task.status is None
        assert do_stuff_task.start_cond == AlwaysTrue()
        assert do_stuff_task.name.endswith(":do_stuff")

        assert [do_stuff_task] == list(session.tasks)

@pytest.mark.parametrize(
    "start_cond,depend,expected",
    [
        pytest.param(
            AlwaysTrue(),
            None,
            AlwaysTrue(),
            id="AlwaysTrue"),
    ],
)
def test_set_start_condition(tmpdir, start_cond, depend, expected, session):

    # Going to tempdir to dump the log files there
    with tmpdir.as_cwd() as old_dir:

        task = FuncTask(
            lambda : None, 
            name="task",
            start_cond=start_cond,
            execution="main",
        )
        assert expected == task.start_cond


@pytest.mark.parametrize(
    "start_cond_str,start_cond",
    [
        pytest.param("true", lambda: AlwaysTrue(), id="true"),
        pytest.param("always true & always true", lambda: AlwaysTrue() & AlwaysTrue(), id="always true & always true"),
    ],
)
def test_set_start_condition_str(tmpdir, start_cond_str, start_cond, session):

    # Going to tempdir to dump the log files there
    with tmpdir.as_cwd() as old_dir:

        task = FuncTask(
            lambda : None, 
            name="task",
            start_cond=start_cond_str,
            execution="main",
        )
        assert start_cond() == task.start_cond

        assert str(task.start_cond) == start_cond_str

@pytest.mark.parametrize(
    "get_task,exc",
    [
        pytest.param(
            lambda: FuncTask(lambda : None, name="task", start_cond="this is not valid", execution="main"), 
            ParserError, 
            id="invalid start_cond"
        ),
        pytest.param(
            lambda: FuncTask(lambda : None, name="task", execution="not valid"), 
            ValueError, 
            id="invalid execution"
        ),
    ],
)
def test_failure(session, exc, get_task):
    with pytest.raises(exc):
        get_task()
    assert session.tasks == set()

def test_rename(session):
    task1 = FuncTask(lambda : None, name="a task 1", execution="main")
    task2 = FuncTask(lambda : None, name="a task 2", execution="main")
    assert session.tasks == {task1, task2}
    assert 'renamed task' not in session
    assert 'a task 1' in session

    task1.name = "renamed task"
    assert task1.name == "renamed task"
    assert session.tasks == {task1, task2}
    assert 'renamed task' in session
    assert 'a task 1' not in session

def test_rename_conflict(session):
    task1 = FuncTask(lambda : None, name="a task 1", execution="main")
    task2 = FuncTask(lambda : None, name="a task 2", execution="main")
    assert session.tasks == {task1, task2}
    assert 'renamed task' not in session
    assert 'a task 1' in session

    with pytest.raises(ValueError):
        task1.name = "a task 2"
    assert session.tasks == {task1, task2}
    assert session['a task 2'] is task2
    assert session['a task 2'] is not task1

def test_existing_raise(session):
    assert session.config.task_pre_exist == 'raise'

    task1 = FuncTask(lambda : None, name="a task", execution="main")
    with pytest.raises(ValueError):
        task2 = FuncTask(lambda : None, name="a task", execution="main")
    assert session.tasks == {task1}

def test_existing_ignore(session):
    session.config.task_pre_exist = 'ignore'
    task1 = FuncTask(lambda : None, name="a task", execution="main")
    task2 = FuncTask(lambda : None, name="a task", execution="main")
    assert session.tasks == {task1}

@pytest.mark.skip(reason="No support for this yet")
def test_existing_replace(session):
    session.config.task_pre_exist = 'replace'
    task1 = FuncTask(lambda : None, name="a task", execution="main")
    task2 = FuncTask(lambda : None, name="a task", execution="main")
    assert session.tasks == {"a task": task2}

def test_existing_rename(session):
    session.config.task_pre_exist = 'rename'
    task1 = FuncTask(lambda : None, name="a task", execution="main")
    task2 = FuncTask(lambda : None, name="a task", execution="main")
    assert session.tasks == {task1, task2}
    assert task2.name == 'a task - 1'

    assert session['a task'] is task1
    assert session['a task - 1'] is task2



from .session import Session
from .core import Scheduler

from ._setup import _setup_defaults
from . import (
    conditions,
    log,
    
    args,
    time,
    tasks,
)
from .tasks import FuncTask
_setup_defaults()
session = Session()
session.set_as_default()

from .application import Rocketry

from . import _version
__version__ = _version.get_versions()['version']
